/*
 *******************************************************************
 *
 * FILE
 *
 * converter.java
 * by Phuong, The NGUYEN
 *
 * a converter
 * as part of a summer scholarship in RSISE - ANU
 * supervised by: Rajeev, Prabhakar GORE
 *
 * begin: 27 Jan 01
 * last updated: 5 Feb 01
 *
 *******************************************************************
 * 
 * WHAT IT DOES
 *
 * converts the result of the parser into the form that is used by
 * the prover
 *
 * Currently it assumes that the input is already in NNF.
 * 
 *******************************************************************
 */


package client;
import card.*;
public class converter
{
	/* all of the node in the parse tree will be numbered from 0 (root)
	 * The numbering is done so that the root is always 0, and nodes in
	 * lower level (larger depth) will be given larger index than that
	 * of nodes in higher levels.
	 * the nature list will indicate the nature of each node. Thus
	 * nature[0] = Constants.DISJ if the root is a disjunction ... etc.
	 */
	
	public byte childs [], nature [];
	public byte num_forms;


	/* convert
	 * it is given a formula parsed by the parser
	 * In the parser, the root is always the last entry in the array
	 * Here we will store the root at 0, and every child of a node will
	 * have an index greater than that node's.
	 * We will traverse the parse tree as given by the parser, collect
	 * nodes and put them in a stack (breath search).
	 *
	 * In the parse tree given by the parser, every node is either AND,
	 * OR, ALL, SOME, ATOM or NEGA. At the leaves we have the value of
	 * the atoms. The child of an node is a leave iff that node is an 
	 * ATOM. 
	 *
	 * Here we have the same kinds of (inner)node: AND, OR, ALL, SOME,
	 * ATOM and NEGA. Also, at leaves we have the values of the atoms.
	 * However, the child of an node is a leave iff that node is either
	 * ATOM or NEGA (Can be thought of as LITERAL). Thus we reduce the
	 * unnecessary NEGA nodes.
	 *
	 * Assuming the given formula is in NNF, then there is no IMPLY,
	 * and after each NEGA there is an atom
	 */
	public void convert(parser par)
	{		
		int i, j, s;
		byte [] childs_tmp = new byte [par.action_obj.num_forms << 1];
		byte [] nature_tmp = new byte [par.action_obj.num_forms];
		
		/* stk_fms: holding the node in the original parse tree
		 * num_forms will be the stack pointer
		 */
		byte [] stk_fms = new byte[par.action_obj.num_forms];
		num_forms = 0;
		
		/* the root is always the last entry in the array representing
		 * the parse tree given by the parser 
		 */
		stk_fms[num_forms++] = (byte)(par.action_obj.num_forms-1);
		
		for(i=0; i<2*par.action_obj.num_forms; i++)
			childs_tmp[i] = -1;
			
		/* rename the node in the parse tree for use of this prover */
		for(i=0; i<par.action_obj.num_forms; i++)
		{
			j = i << 1; s = stk_fms[i] << 1;
			nature_tmp[i] = par.action_obj.nature[stk_fms[i]];
			
			/* if the current node is an NEGA, then its child is
			 * an ATOM whose child is a leave with value of that atom
			 */
			if(nature_tmp[i] == card.proverInterface.NEGA)
			{
				childs_tmp[j] = 
				par.action_obj.childs[par.action_obj.childs[s] << 1];
			}
			
			/* if the child is an atom, then copy the value of the atom
			 * into the value of the current node
			 */
			else if(nature_tmp[i] == card.proverInterface.PROP)
				childs_tmp[j] = par.action_obj.childs[s];
				
			/* if the child is not an ATOM or a NEGA, the stack pointer
			 * will be used to number the (first) child of the current
			 * node, and we need to push this child in the stack
			 */
			else
			{
				childs_tmp[j] = num_forms;
				stk_fms[num_forms++] = par.action_obj.childs[s];
			}
			
			/* if this node has two childen then copy the second child
			 * and also put it in the stack
			 */
			if(nature_tmp[i] == card.proverInterface.CONJ || 
				nature_tmp[i] == card.proverInterface.DISJ)
			{
				childs_tmp[j | 1] = num_forms;
				stk_fms[num_forms++] = par.action_obj.childs[s | 1];
			}
		}
		
		/* allocate the childs and nature */
		childs = new byte[num_forms << 1];
		nature = new byte[num_forms];
		System.arraycopy(childs_tmp,0,childs,0,num_forms << 1);
		System.arraycopy(nature_tmp,0,nature,0,num_forms);
	}

	public void display()
	{
		System.out.println("\n---------- CONVERTER CLASS -------------");
		System.out.println("num_forms = " + num_forms);

        for(int i=0; i<num_forms; i++)
        {
            System.out.print(i + ")\t" + nodeName(nature[i])+"\t");

            if(nature[i] == card.proverInterface.PROP)
            {
            	if(childs[2*i] == 0)
            	{
            		System.out.print(childs[2*i] + "\t" + childs[2*i+1] + "\t\t(ATOM TRUE)");
            	}
            	else System.out.print(childs[2*i] + "\t" + childs[2*i+1] + "\t\t(ATOM " + childs[2*i] + ")");
            }
            else if(nature[i] == card.proverInterface.NEGA)
            {
            	if(childs[2*i] == 0)
            	{
            		System.out.print(childs[2*i] + "\t" + childs[2*i+1] + "\t\t(NEGA ATOM TRUE)");
            	}
            	else System.out.print(childs[2*i] + "\t" + childs[2*i+1] + "\t\t(NEGA ATOM " + childs[2*i] + ")");
            }            	
            else if(nature[i] <= card.proverInterface.SOME)
            {
            	System.out.print(childs[2*i] + "\t" + childs[2*i+1]);
            }
            else System.out.print("Unrecognised child");
            System.out.println();
        }
        System.out.println("********************************");
    }

    public String nodeName(byte b)
    {
        if(b == card.proverInterface.DISJ) return "DISJ";
        else if(b == card.proverInterface.CONJ) return "CONJ";
        else if(b == card.proverInterface.NEGA) return "NEGA";
        else if(b == card.proverInterface.PROP) return "PROP";
        else if(b == card.proverInterface.SOME) return "SOME";
        else if(b == card.proverInterface.ALL) return "ALL";
        else return "UNKNOWN";
    }
}